import os
import sys
from typing import List


RAKAM_SYSTEMS_DIR = os.path.dirname(
    os.path.dirname(os.path.abspath(__file__))  # ingestion  # this file
)
sys.path.append(RAKAM_SYSTEMS_DIR)

from rakam_systems.core import VSFile, NodeMetadata, Node


def llama_documents_to_VSFiles(llama_documents) -> List[VSFile]:
    """
    Convert a list of LlamaIndex documents (from different sources) to a list of VSFiles.
    """
    vs_files = []
    page_number_tracker = 1
    current_file_name = None

    for doc in llama_documents:
        file_name = doc.metadata["file_name"]
        if file_name != current_file_name:  # new file
            current_file_name = file_name
            page_number_tracker = 1
            vs_file = VSFile(file_name)  # create a new VSFile
            vs_files.append(vs_file)

        # create a new node
        node_metadata = NodeMetadata(
            source_file_uuid=vs_file.uuid, position=page_number_tracker
        )
        node = Node(doc.text, node_metadata)
        vs_file.nodes.append(node)
        page_number_tracker += 1

    return vs_files


def llama_documents_to_VSFile(llama_documents) -> VSFile:
    """
    Convert a list of LlamaIndex documents (from the same source) to a VSFile.
    """
    file_name = llama_documents[0].metadata["file_name"]
    vs_file = VSFile(file_name)
    page_number_tracker = 1

    for doc in llama_documents:
        node_metadata = NodeMetadata(
            source_file_uuid=vs_file.uuid, position=page_number_tracker
        )
        node = Node(doc.text, node_metadata)
        vs_file.nodes.append(node)
        page_number_tracker += 1

    return vs_file


def parsed_url_to_VSFile(
    url: str, extracted_content: str, other_info: dict = None
) -> VSFile:
    """
    Convert a parsed URL to a VSFile.
    """
    vs_file = VSFile(file_path=url)
    node_metadata = NodeMetadata(
        source_file_uuid=vs_file.uuid,
        position=0,  # what is position for urls?
        custom=other_info,
    )
    node = Node(content=extracted_content, metadata=node_metadata)
    vs_file.nodes = [node]  # All content is in one node before processing
    return vs_file



import logging
import os
import pickle
import time
from typing import Any
from typing import Dict
from typing import List

import faiss
import numpy as np
from sentence_transformers import SentenceTransformer

from rakam_systems.core import VSFile

# Configure logging
logging.basicConfig(level=logging.INFO)


class VectorStores:
    """
    A class for managing multiple vector stores using FAISS and SentenceTransformers.
    """

    def __init__(self, base_index_path: str, embedding_model: str) -> None:
        """
        Initializes the VectorStores with the specified base index path and embedding model.

        :param base_index_path: Base path to store the FAISS indexes.
        :param embedding_model: Pre-trained SentenceTransformer model name.
        """
        self.base_index_path = base_index_path
        if not os.path.exists(self.base_index_path):
            os.makedirs(self.base_index_path)

        self.embedding_model = SentenceTransformer(embedding_model)
        self.stores = {}
        self.load_all_stores()

    def load_all_stores(self) -> None:
        """
        Loads all vector stores from the base directory.
        """
        for store_name in os.listdir(self.base_index_path):
            store_path = os.path.join(self.base_index_path, store_name)
            if os.path.isdir(store_path):
                self.stores[store_name] = self.load_store(store_path)

    def load_store(self, store_path: str) -> Dict[str, Any]:
        """
        Loads a single vector store from the specified directory.

        :param store_path: Path to the store directory.
        :return: Dictionary containing the store's index, nodes, and metadata.
        """
        store = {}
        try:
            store["index"] = faiss.read_index(os.path.join(store_path, "index"))
            with open(
                os.path.join(store_path, "category_index_mapping.pkl"), "rb"
            ) as f:
                store["category_index_mapping"] = pickle.load(f)
            with open(
                os.path.join(store_path, "metadata_index_mapping.pkl"), "rb"
            ) as f:
                store["metadata_index_mapping"] = pickle.load(f)
            with open(os.path.join(store_path, "nodes.pkl"), "rb") as f:
                store["nodes"] = pickle.load(f)
            logging.info(f"Store loaded successfully from {store_path}.")
        except Exception as e:
            logging.warning(f"Error loading store from {store_path}: {e}")
        return store

    def predict_embeddings(self, query: str) -> np.ndarray:
        """
        Predicts embeddings for a given query using the embedding model.

        :param query: Query string to encode.
        :return: Embedding vector for the query.
        """
        logging.info(f"Predicting embeddings for query: {query}")
        query_embedding = self.embedding_model.encode(query)
        query_embedding = np.asarray([query_embedding], dtype="float32")
        return query_embedding

    def search(
        self, store_name: str, query: str, distance_type="cosine", number=5
    ) -> dict:
        """
        Searches the specified store for the closest embeddings to the query.

        :param store_name: Name of the store to search in.
        :param query: Query string to search for.
        :param distance_type: Distance metric to use (default is "cosine").
        :param number: Number of closest embeddings to return.
        :return: Dictionary of search results with suggestion texts and distances.
        """
        logging.info(
            f"Searching in store: {store_name} for query: {query} with distance type: {distance_type} and number of results: {number}"
        )

        store = self.stores.get(store_name)
        if not store:
            raise ValueError(f"No store found with name: {store_name}")

        query_embedding = self.predict_embeddings(query)
        suggested_nodes = []

        if distance_type == "cosine":
            faiss.normalize_L2(query_embedding)
        D, I = store["index"].search(query_embedding, number)

        seen_texts = set()
        valid_suggestions = {}
        count = 0
        for i, id_ in enumerate(I[0]):
            if count >= number:
                break
            if id_ != -1 and id_ in store["category_index_mapping"]:
                suggestion_text = store["category_index_mapping"][id_]
                node_metadata = store["metadata_index_mapping"][id_]
                suggested_nodes.append(store["nodes"][id_])
                if suggestion_text not in seen_texts:
                    seen_texts.add(suggestion_text)
                    valid_suggestions[str(id_)] = (
                        node_metadata,
                        suggestion_text,
                        float(D[0][i]),
                    )
                    count += 1

        while count < number:
            valid_suggestions[f"placeholder_{count}"] = (
                "No suggestion available",
                float("inf"),
            )
            count += 1

        logging.info(f"Search results: {valid_suggestions}")
        return valid_suggestions, suggested_nodes

    def get_embeddings(self, sentences: List[str], parallel: bool = True) -> np.ndarray:
        """
        Generates embeddings for a list of sentences.

        :param sentences: List of sentences to encode.
        :param parallel: Whether to use parallel processing (default is True).
        :return: Embedding vectors for the sentences.
        """
        logging.info(f"Generating embeddings for {len(sentences)} sentences.")
        start = time.time()
        if parallel:
            os.environ["TOKENIZERS_PARALLELISM"] = "false"
            pool = self.embedding_model.start_multi_process_pool(
                target_devices=["cpu"] * 5
            )
            embeddings = self.embedding_model.encode_multi_process(
                sentences, pool, batch_size=16
            )
            self.embedding_model.stop_multi_process_pool(pool)
        else:
            os.environ["TOKENIZERS_PARALLELISM"] = "true"
            embeddings = self.embedding_model.encode(
                sentences,
                batch_size=32,
                show_progress_bar=True,
                convert_to_tensor=True,
            )
        logging.info(
            f"Time taken to encode {len(sentences)} items: {round(time.time() - start, 2)} seconds"
        )
        return embeddings.cpu().detach().numpy()

    def create_from_files(self, stores_files: Dict[str, List[VSFile]]) -> None:
        """
        Creates FAISS indexes from dictionaries of store names and VSFile objects.

        :param stores_files: Dictionary where keys are store names and values are lists of VSFile objects.
        """
        for store_name, files in stores_files.items():
            logging.info(f"Creating FAISS index for store: {store_name}")
            text_chunks = []
            metadata = []
            nodes = []

            for file in files:
                for node in file.nodes:
                    nodes.append(node)
                    text_chunks.append(node.content)
                    formatted_metadata = {
                        "node_id": node.metadata.node_id,
                        "source_file_uuid": node.metadata.source_file_uuid,
                        "position": node.metadata.position,
                        "custom": node.metadata.custom,
                    }
                    metadata.append(formatted_metadata)

            self._create_and_save_index(store_name, nodes, text_chunks, metadata)

    def create_from_nodes(self, store_name: str, nodes: List[Any]) -> None:
        """
        Creates a FAISS index from a list of nodes and stores it under the given store name.

        :param store_name: Name of the store to create.
        :param nodes: List of nodes containing the content and metadata.
        """
        logging.info(f"Creating FAISS index for store: {store_name}")
        text_chunks = []
        metadata = []

        for node in nodes:
            text_chunks.append(node.content)
            formatted_metadata = {
                "node_id": node.metadata.node_id,
                "source_file_uuid": node.metadata.source_file_uuid,
                "position": node.metadata.position,
                "custom": node.metadata.custom,
            }
            metadata.append(formatted_metadata)

        self._create_and_save_index(store_name, nodes, text_chunks, metadata)

    def _create_and_save_index(
        self,
        store_name: str,
        nodes: List[Any],
        text_chunks: List[str],
        metadata: List[Dict[str, Any]],
    ) -> None:
        """
        Helper function to create and save a FAISS index.

        :param store_name: Name of the store to create.
        :param nodes: List of nodes.
        :param text_chunks: List of text chunks to encode and index.
        :param metadata: List of metadata associated with the text chunks.
        """
        store_path = os.path.join(self.base_index_path, store_name)
        if not os.path.exists(store_path):
            os.makedirs(store_path)

        data_embeddings = self.get_embeddings(sentences=text_chunks, parallel=False)
        category_index_mapping = dict(zip(range(len(text_chunks)), text_chunks))

        # Save category index mapping to file
        with open(os.path.join(store_path, "category_index_mapping.pkl"), "wb") as f:
            pickle.dump(category_index_mapping, f)

        # Save nodes to file
        with open(os.path.join(store_path, "nodes.pkl"), "wb") as f:
            pickle.dump(nodes, f)

        index = faiss.IndexIDMap(faiss.IndexFlatIP(data_embeddings.shape[1]))
        faiss.normalize_L2(data_embeddings)
        index.add_with_ids(
            data_embeddings, np.array(list(category_index_mapping.keys()))
        )
        faiss.write_index(index, os.path.join(store_path, "index"))

        metadata_index_mapping = dict(zip(range(len(text_chunks)), metadata))

        # Save metadata nodes to file
        with open(os.path.join(store_path, "metadata_index_mapping.pkl"), "wb") as f:
            pickle.dump(metadata_index_mapping, f)

        self.stores[store_name] = {
            "index": index,
            "nodes": nodes,
            "category_index_mapping": category_index_mapping,
            "metadata_index_mapping": metadata_index_mapping,
        }
        logging.info(
            f"FAISS index for store {store_name} created and saved successfully."
        )



import glob
import json
import os
import re
import sys
from abc import ABC
from abc import abstractmethod
from typing import List
from typing import Type

import dotenv
import joblib
import pymupdf
import requests
from bs4 import BeautifulSoup
from llama_parse import LlamaParse
from playwright.sync_api import sync_playwright

dotenv.load_dotenv()

import logging

from rakam_systems.ingestion import utils


RAKAM_SYSTEMS_DIR = os.path.dirname(
    os.path.dirname(os.path.abspath(__file__))  # ingestion  # this file
)
sys.path.append(RAKAM_SYSTEMS_DIR)

from rakam_systems.core import VSFile, NodeMetadata, Node


class ContentExtractor(ABC):
    """
    Interface for content extractors.

    Attributes:
        output_format (str): The desired output format of the extracted content

    Methods:
        extract_content: Fetch and parse content from the source
        format_content: Format the parsed content before returning it
    """

    def __init__(self, output_format: str = "markdown"):
        self.output_format = output_format

    @abstractmethod
    def extract_content(self, source):
        pass

    def format_content(self, content):
        # Optional
        # To format the content before returning it, override this method.
        return content


### --- PDFs --- ###
class SimplePDFParser:
    """
    A simple PDF parser that extracts text from PDF files using PyMuPDF.
    """

    def __init__(self) -> None:
        pass

    def parse_one_pdf(self, file_path: str):
        try:
            doc = pymupdf.open(file_path)
        except Exception as e:
            logging.error(f"Error reading PDF file: {file_path}")
            return None

        vs_file = VSFile(file_path)  # create a new VSFile
        nodes = []
        for page_num, page in enumerate(doc):
            text = page.get_text().strip()
            text = self._clean_text(text)
            if text:
                metadata = NodeMetadata(
                    source_file_uuid=file_path, position=(page_num + 1)
                )
                node = Node(text, metadata)
                nodes.append(node)
        vs_file.nodes = nodes
        return vs_file

    def _clean_text(text):
        # Remove non-printable and control characters
        text = re.sub(r"[\x00-\x1F\x7F-\x9F]", "", text)
        # Normalize whitespace
        text = re.sub(r"\s+", " ", text).strip()
        return text


class LlamaPDFParser:
    """
    A PDF parser that extracts text from PDF files using LlamaParse.

    Methods:
        parse_one_pdf: Parse a single PDF file and return a VSFile
    """

    def __init__(self, persist: bool = True) -> None:
        self.parser = LlamaParse(
            api_key=os.getenv("LLAMA_CLOUD_API_KEY"),
            result_type="text",
            num_workers=4,
            show_progress=True,
            language="fr",
            parsing_instruction="",
            split_by_page=True,
        )
        self.persist = persist

    def parse_one_pdf(self, file_path: str):
        with open(file_path, "rb") as f:
            extra_info = {"file_name": file_path}
            documents = self.parser.load_data(f, extra_info=extra_info)

        if self.persist:
            persist_dir = os.path.dirname(file_path) + "_parsed"
            os.makedirs(persist_dir, exist_ok=True)
            for page_n, doc in enumerate(documents):
                file_name = os.path.basename(file_path)
                output_path = os.path.join(
                    persist_dir, file_name.replace(".pdf", f"_p{page_n+1}.pkl")
                )
                joblib.dump(doc, output_path)

        VSFile = utils.llama_documents_to_VSFile(documents)
        return VSFile


class PDFContentExtractor(ContentExtractor):
    """
    Extracts content from PDF files.

    Attributes:
        output_format (str): The desired output format of the extracted content
        parser_name (str): The name of the parser to use (default: "LlamaParse")

        persist (bool): Persist the parsed documents (default: True) (LlamaParse only)

    Methods:
        get_parser: Initialize the parser
        extract_content: Parse content from the PDFs (pass path to file or directory)
    """

    def __init__(self, parser_name: str = "LlamaParse", persist: bool = True, **kwargs):
        super().__init__(**kwargs)
        self.parser_name = parser_name
        self.persist = persist
        self.parser = self._get_parser(parser_name, persist=persist)

    def _get_parser(self, parser_name: str, persist: bool):
        if parser_name == "LlamaParse":
            return LlamaPDFParser(persist=persist)
        elif parser_name == "SimplePDFParser":
            return SimplePDFParser()
        else:
            raise NotImplementedError(
                f"Unsupported parser: {parser_name}. Use 'LlamaParse' or 'SimplePDFParser'."
            )

    def _load_parsed_files(self, parsed_files) -> VSFile:
        """
        Load previously parsed files (all pkls for a PDF).
        """
        documents = []
        for file_path in parsed_files:
            doc = joblib.load(file_path)
            documents.append(doc)
        VSFile = utils.llama_documents_to_VSFile(documents)
        return VSFile

    def _load_or_parse_pdf(self, file_path) -> VSFile:
        """
        Determine whether to load or parse a PDF file.
        """
        filename = os.path.basename(file_path)
        parsed_dir = os.path.dirname(file_path) + "_parsed"
        if os.path.exists(parsed_dir):
            parsed_files = glob.glob(
                os.path.join(parsed_dir, filename.replace(".pdf", "_p*.pkl"))
            )
            if parsed_files:
                return self._load_parsed_files(parsed_files)
        return self.parser.parse_one_pdf(file_path)

    def _parse_directory(self, directory) -> List[VSFile]:
        """
        Parse all PDF files in a directory.
        """
        VSFiles = []
        for filename in os.listdir(directory):
            if filename.lower().endswith(".pdf"):
                file_path = os.path.join(directory, filename)
                VSFile = self._load_or_parse_pdf(file_path)
                VSFiles.append(VSFile)
        return VSFiles

    def extract_content(self, source) -> List[VSFile]:
        """
        Wrapper method to extract content from PDF files.
        """
        if not os.path.exists(source):
            raise FileNotFoundError(f"Source not found: {source}")

        if os.path.isdir(source):
            VSFiles = self._parse_directory(source)
        elif os.path.isfile(source) and source.lower().endswith(".pdf"):
            VSFile = self._load_or_parse_pdf(source)
            VSFiles = [VSFile]
        else:
            raise ValueError(
                "Source must be a PDF file or a directory containing PDF files."
            )
        return VSFiles


### --- URLs --- ###
class URLContentExtractor(ContentExtractor):
    """
    Extracts content from URLs. Returns a VSFile with the extracted content.

    Attributes:
        output_format (str): The desired output format of the extracted content
        scraper (str): The name of the scraper to use (default: "BeautifulSoup")

    Methods:
        extract_content: Fetch and parse content from the URL
        format_content: Format the parsed content before returning it as a VSFile
    """

    def __init__(self, scraper: str = "BeautifulSoup", **kwargs):
        super().__init__(**kwargs)
        self.scraper = scraper
        self.soup = None
        self.unique_lines = set()

    def extract_content(self, source):
        if self.scraper == "BeautifulSoup":
            response = requests.get(source)
            response.raise_for_status()
            self.soup = BeautifulSoup(response.content, "html.parser")
        elif self.scraper == "Playwright":
            with sync_playwright() as p:
                browser = p.chromium.launch()
                page = browser.new_page()
                page.goto(source)
                html = page.content()
                browser.close()
            self.soup = BeautifulSoup(html, "html.parser")
        else:
            raise NotImplementedError(
                f"Unsupported scraper: {self.scraper}. Use 'BeautifulSoup' or 'Playwright'."
            )

        return self.format_content(self.soup, source)

    def format_content(self, content, source):
        if self.output_format == "markdown":
            md_content = self.convert_to_markdown(content)
            # Create VSFile
            return utils.parsed_url_to_VSFile(url=source, extracted_content=md_content)
        elif self.output_format == "text":
            # Create VSFile
            return utils.parsed_url_to_VSFile(url=source, extracted_content=content)

    def convert_to_markdown(self):
        markdown_content = ""
        for element in self.soup.find_all(True):  # All tags
            markdown_content += self._extract_text(element)
        return markdown_content

    def _extract_text(self, element):
        # Markdown conversion logic
        # Headings
        if element.name in ["h1", "h2", "h3", "h4", "h5", "h6"]:
            prefix = "#" * int(element.name[1])
            line = f"\n{prefix} {element.get_text().strip()}\n\n"
            if line not in self.unique_lines:
                self.unique_lines.add(line)
                return line
            return ""

        # Paragraphs
        elif element.name == "p":
            line = f"{element.get_text().strip()}\n\n"
            if line not in self.unique_lines:
                self.unique_lines.add(line)
                return line
            return ""

        # Lists
        elif element.name == "li":
            line = f"- {element.get_text().strip()}\n"
            if line not in self.unique_lines:
                self.unique_lines.add(line)
                return line
            return ""
        elif element.name == "ul":
            return "".join(
                [self.extract_text(child) for child in element.find_all("li")]
            )
        elif element.name == "ol":
            return "".join(
                [
                    f"{idx+1}. {child.get_text().strip()}\n"
                    for idx, child in enumerate(element.find_all("li"))
                ]
            )

        # Images
        elif element.name == "img":
            line = f"![{element.get('alt', '')}]({element.get('src', '')})\n\n"
            if line not in self.unique_lines:
                self.unique_lines.add(line)
                return line
            return ""

        # Links
        elif element.name == "a":
            line = f"[{element.get_text().strip()}]({element.get('href', '')})"
            if line not in self.unique_lines:
                self.unique_lines.add(line)
                return line
            return ""

        # Tables
        elif element.name == "table":
            rows = element.find_all("tr")
            headers = rows[0].find_all("th")
            header_row = (
                "| "
                + " | ".join([header.get_text().strip() for header in headers])
                + " |\n"
            )
            separator_row = "| " + " | ".join(["---" for _ in headers]) + " |\n"
            body_rows = [
                "| "
                + " | ".join([td.get_text().strip() for td in row.find_all("td")])
                + " |\n"
                for row in rows[1:]
            ]
            table_text = header_row + separator_row + "".join(body_rows) + "\n\n"
            if table_text not in self.unique_lines:
                self.unique_lines.add(table_text)
                return table_text
            return ""

        # Text formatting
        elif element.name in ["strong", "b"]:
            line = f"**{element.get_text().strip()}**"
            if line not in self.unique_lines:
                self.unique_lines.add(line)
                return line
            return ""
        elif element.name in ["em", "i"]:
            line = f"*{element.get_text().strip()}*"
            if line not in self.unique_lines:
                self.unique_lines.add(line)
                return line
            return ""

        else:
            # If the element has child elements, recursively extract their text
            if element.find_all(recursive=False):
                child_texts = "".join(
                    [
                        self.extract_text(child)
                        for child in element.find_all(recursive=False)
                    ]
                )
                return child_texts
            else:
                # If the element has no child elements, return its stripped text
                line = element.get_text().strip()
                if line and line not in self.unique_lines:
                    self.unique_lines.add(line)
                    return line
                return ""


### --- JSONs --- ###
class JSONContentExtractor(ContentExtractor):
    """
    Extracts content from JSON files.

    Methods:
        extract_content: Load and parse content from the JSON file(s)
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def _load_one_json(self, file_path):
        with open(file_path, "r") as f:
            data = json.load(f)
            content = data.get("content", "")
            if not content:
                raise ValueError("No content found in the JSON file.")

            # all other keys become custom metadata
            other_info = {key: value for key, value in data.items() if key != "content"}

            vs_file = VSFile(file_path=file_path)
            node_metadata = NodeMetadata(
                source_file_uuid=vs_file.uuid, position=0, custom=other_info
            )
            node = Node(content, node_metadata)
            vs_file.nodes = [node]
            return vs_file

    def extract_content(self, source):
        if os.path.isfile(source) and source.lower().endswith(".json"):
            vs_file = self._load_one_json(source)
            return [vs_file]
        elif os.path.isdir(source):
            vs_files = []
            for filename in os.listdir(source):
                if filename.lower().endswith(".json"):
                    file_path = os.path.join(source, filename)
                    vs_file = self._load_one_json(file_path)
                    vs_files.append(vs_file)
            return vs_files
        else:
            raise ValueError(
                "Source must be a JSON file or a directory containing JSON files."
            )


if __name__ == "__main__":  # example usage
    # one file, vs_files will be a list with one element
    pdf_extractor = PDFContentExtractor(
        parser_name="SimplePDFParser", output_format="markdown"
    )
    vs_files = pdf_extractor.extract_content(source="example.pdf")

    # directory
    pdf_extractor = PDFContentExtractor(parser_name="SimplePDFParser")
    vs_files = pdf_extractor.extract_content(source="path/to/pdf_directory")

    url_extractor = URLContentExtractor(
        scraper="BeautifulSoup", output_format="markdown"
    )
    vs_file = url_extractor.extract_content(source="https://www.example.com")

    json_extractor = JSONContentExtractor()
    vs_files = json_extractor.extract_content(source="example.json")



import logging
import os
from typing import Any
from typing import Dict
from typing import List
from typing import Optional

import dotenv
from openai import OpenAI

logging.basicConfig(level=logging.INFO)


class LLM:
    def __init__(self, model: str, api_key: str = None) -> None:
        self.model = model
        self.client = OpenAI(
            api_key=os.getenv("OPENAI_API_KEY") if not api_key else api_key
        )

    def call_llm(self, sys_prompt: str, prompt: str) -> str:
        completion = self.client.chat.completions.create(
            model=self.model,
            messages=[
                {"role": "system", "content": sys_prompt},
                {"role": "user", "content": prompt},
            ],
        )
        return completion.choices[0].message.content

    def call_llm_stream(self, sys_prompt: str, prompt: str):
        completion = self.client.chat.completions.create(
            model=self.model,
            messages=[
                {"role": "system", "content": sys_prompt},
                {"role": "user", "content": prompt},
            ],
            stream=True,
        )

        for chunk in completion:
            yield chunk.choices[0].delta.content



import logging
from abc import ABC
from abc import abstractmethod
from typing import Any

import dotenv
import pandas as pd

from rakam_systems.core import Node
from rakam_systems.core import NodeMetadata
from rakam_systems.generation.llm import LLM
from rakam_systems.vector_store import VectorStores


logging.basicConfig(level=logging.INFO)
dotenv.load_dotenv()


# Abstract Action class
class Action(ABC):
    @abstractmethod
    def __init__(self, agent, **kwargs):
        pass

    @abstractmethod
    def execute(self, **kwargs):
        pass


class ClassifyQuery(Action):
    def __init__(
        self,
        agent,
        trigger_queries: pd.Series,
        class_names: pd.Series,
        embedding_model: str = "all-MiniLM-L6-v2",
    ):
        self.agent = agent
        self.trigger_queries = trigger_queries
        self.class_names = class_names
        self.embedding_model = embedding_model
        self.vector_store = self.build_vector_store(trigger_queries, class_names)

    def build_vector_store(
        self, trigger_queries: pd.Series, class_names: pd.Series
    ) -> VectorStores:
        """
        Builds a VectorStores object from the trigger queries and class names.
        """
        # Create nodes from trigger queries and class names
        nodes = []
        for query, class_name in zip(trigger_queries, class_names):
            metadata = NodeMetadata(
                source_file_uuid=query, position=None, custom={"class_name": class_name}
            )
            node = Node(content=query, metadata=metadata)
            nodes.append(node)

        # Initialize VectorStores
        vector_store = VectorStores(
            base_index_path="temp_path", embedding_model=self.embedding_model
        )

        # Use the create_from_nodes method to build the index
        vector_store.create_from_nodes(store_name="query_classification", nodes=nodes)

        return vector_store

    def execute(self, query: str):
        """
        Classifies the query by finding the closest match in the FAISS index.
        """
        # Perform the search using the vector store
        _, node_search_results = self.vector_store.search(
            store_name="query_classification", query=query, number=2
        )

        # Extract the matched trigger query and class name
        matched_node = node_search_results[0]
        matched_trigger_query = matched_node.content
        matched_class_name = matched_node.metadata.custom["class_name"]

        print("Successfully classified query:", query)
        return matched_class_name, matched_trigger_query


class PromptGeneration(Action):
    def __init__(
        self,
        agent,
        sys_prompt: str,
        prompt: str,
    ):
        self.agent = agent
        self.sys_prompt = sys_prompt
        self.prompt = prompt

    def execute(self, query, prompt_kwargs: dict = {}, stream: bool = False):
        ### --- Format Prompt --- ###
        formatted_prompt = self.prompt.format(
            query=query,
            **prompt_kwargs,
        )

        ### --- LLM Generation --- ###
        if stream:
            return self._generate_stream(formatted_prompt)
        else:
            return self._generate_non_stream(formatted_prompt)

    # STREAMING (returns a generator)
    def _generate_stream(self, formatted_prompt):
        # Call the LLM to generate the final answer in streaming mode
        response_generator = self.agent.llm.call_llm_stream(
            self.sys_prompt, formatted_prompt
        )
        for chunk in response_generator:
            yield chunk

    # NON-STREAMING (returns a string)
    def _generate_non_stream(self, formatted_prompt) -> str:
        # Call the LLM to generate the final answer
        answer = self.agent.llm.call_llm(self.sys_prompt, formatted_prompt)
        return answer


class RAGGeneration(Action):
    def __init__(
        self,
        agent,
        sys_prompt: str,
        prompt: str,
        vector_stores: VectorStores,
        vs_descriptions: dict = None,
    ):
        self.agent = agent
        self.sys_prompt = sys_prompt
        self.prompt = prompt
        self.vector_stores = vector_stores
        self.vs_descriptions = vs_descriptions

        self.store_separator = "\n====\n"
        self.result_separator = "\n----\n"

    def execute(self, query, prompt_kwargs: dict = {}, stream: bool = False):
        ### --- Vector Store Search --- ###
        formatted_search_results = self._perform_vector_store_search(query)

        ### --- Format Prompt --- ###
        formatted_prompt = self.prompt.format(
            query=query,
            search_results=formatted_search_results,
            **prompt_kwargs,
        )

        ### --- LLM Generation --- ###
        if stream:
            return self._generate_stream(formatted_prompt)
        else:
            return self._generate_non_stream(formatted_prompt)

    # STREAMING (returns a generator)
    def _generate_stream(self, formatted_prompt):
        # Call the LLM to generate the final answer in streaming mode
        response_generator = self.agent.llm.call_llm_stream(
            self.sys_prompt, formatted_prompt
        )
        for chunk in response_generator:
            yield chunk

    # NON-STREAMING (returns a string)
    def _generate_non_stream(self, formatted_prompt) -> str:
        # Call the LLM to generate the final answer
        answer = self.agent.llm.call_llm(self.sys_prompt, formatted_prompt)
        return answer

    def _perform_vector_store_search(self, query: str) -> str:
        """
        Perform a search across all vector stores and format the results.
        """
        search_results = {}
        for store_name, _ in self.vector_stores.stores.items():
            _, node_search_results = self.vector_stores.search(
                store_name=store_name, query=query
            )
            if node_search_results:
                # Format the search results for one vector store
                search_results[store_name] = self._format_search_results(
                    node_search_results
                )

        # Format the search results for all vector stores
        formatted_search_results = []
        for store_name, results in search_results.items():
            if store_name in self.vs_descriptions:
                description = self.vs_descriptions[store_name]
            else:
                description = store_name
            formatted_search_results.append(f"\n{description}\n")
            formatted_search_results.append(results)
            formatted_search_results.append(self.store_separator)

        return "".join(formatted_search_results)

    def _format_search_results(self, search_results: list) -> str:
        """
        Formats the search results from one vector store into a string.
        """
        if not search_results:
            return f"No results found. {self.result_separator}"

        formatted_results = [
            f"{node.content}{self.result_separator}" for node in search_results
        ]

        return "".join(formatted_results)


class Agent(ABC):
    def __init__(self, model: str, api_key: str):
        self.llm = LLM(
            model,
            api_key,
        )  # Initialize the LLM with the specified model
        self.state = {}  # Initialize an empty state dictionary
        self.actions = {}  # Dictionary to store actions by name

    def process_state(self, **kwargs) -> dict:
        """
        Build a temporary state based on the input query.

        :param input: The input query string.
        :return: A dictionary representing the state.
        """
        # Example: Building a state based on keyword analysis
        state = {}
        return state

    def add_action(self, action_name: str, action: Action):
        """
        Adds an action to the agent's set of available actions.

        :param action_name: A string name for the action (e.g., 'rag_generation').
        :param action: An instance of an Action subclass.
        """
        self.actions[action_name] = action

    @abstractmethod
    def choose_action(self, input: str, state: Any) -> Action:
        """
        Abstract method to select an action based on input.

        :param input: A string input based on which the action is selected.
        :return: The selected Action instance.
        """
        pass

    def execute_action(self, action_name: str, **kwargs):
        """
        Executes the selected action with the provided arguments.

        :param action_name: The name of the action to execute.
        :param kwargs: Arguments to pass to the action's execute method.
        :return: The result of the action's execute method.
        """
        action = self.choose_action(action_name)
        return action.execute(**kwargs)
